from dew.args import ArgumentData
from dew.dewfile import ProjectFilesParser
from dew.projectproperties import ProjectPropertiesController
from dew.storage import StorageController
from dew.view import View


class CommandData(object):
    def __init__(self, args: ArgumentData, view: View, storage: StorageController,
                 properties: ProjectPropertiesController, project_parser: ProjectFilesParser):
        self.args = args
        self.view = view
        self.storage = storage
        self.properties = properties
        self.project_parser = project_parser



import posixpath

from typing import Optional, Tuple

import git.repo


def fetch(origin: git.Remote, refspec: Optional[str] = None) -> None:
    need_to_fetch_all = False

    try:
        origin.fetch(refspec=refspec)
    except git.GitCommandError as e:
        if 'unadvertised object' in e.stderr:
            # The ref might have been an unadvertised object (e.g. github does not advertise individual commits and
            # does not allow fetching an individual commit). If this is the case, we must fetch everything.
            need_to_fetch_all = True
        else:
            raise

    if need_to_fetch_all:
        origin.fetch()


def checkout(repo: git.Repo, origin: git.Remote, head_name: str, ref: str) -> None:
    if not repo.head.is_valid() or repo.head.commit.hexsha != ref:
        head = repo.create_head(path=head_name, commit=ref)

        tracking_ref = None
        for remote_ref in origin.refs:
            if remote_ref.remote_head == head_name:
                tracking_ref = remote_ref
                break

        head.set_tracking_branch(tracking_ref)
        head.checkout()

    for submodule in repo.submodules:
        print(f'Processing submodule {submodule.name}', flush=True)

        # Calling git directly for own submodules since using relative path is not working in gitpython
        # see https://github.com/gitpython-developers/GitPython/issues/730
        if submodule.url[0:3] == '../':
            repo.git.submodule('init', submodule.name)
            repo.git.submodule('update', submodule.name)
        # For external submodules we can use the update function of gitpython
        else:
            submodule.update(init=True)


def get_repo(url: str, destination_dir: str) -> Tuple[git.Repo, git.Remote]:
    repo: git.Repo = None
    origin: git.Remote = None
    try:
        repo = git.repo.Repo(destination_dir)
    except (git.NoSuchPathError, git.InvalidGitRepositoryError):
        pass

    if repo is None:
        repo = git.repo.Repo.init(path=destination_dir, mkdir=True)

    for remote in repo.remotes:
        if remote.name == 'origin':
            origin = remote

    if origin is None:
        origin = repo.create_remote('origin', url)
    else:
        invalid_urls = set()
        for current_url in origin.urls:
            if current_url != url:
                invalid_urls.add(current_url)
        for invalid_url in invalid_urls:
            origin.set_url(url, invalid_url)

    return repo, origin


def get_latest_ref(origin: git.Remote, head_name: str) -> str:
    ref: git.RemoteReference = None

    for head_ref in origin.refs:
        if head_ref.remote_head == head_name:
            ref = head_ref
            break

    if ref is None:
        raise ValueError('Could not find remote head!')

    return ref.commit.hexsha



from typing import Dict, List, Optional


class DependencyGraphNode(object):
    def __init__(self, name: str):
        self.name = name
        self.children: List[DependencyGraphNode] = []
        self.parent: DependencyGraphNode = None


class DependencyGraph(object):
    def __init__(self):
        self.root = DependencyGraphNode('')
        self.nodes: Dict[str, DependencyGraphNode] = {}

    def add_dependency(self, name: str, parent_name: Optional[str]):
        node = self.nodes.get(name)
        if node is None:
            node = DependencyGraphNode(name)
            self.nodes[name] = node

        parent_node = self.root
        if parent_name:
            parent_node = self.nodes[parent_name]

        parent_node.children.append(node)
        node.parent = parent_node

    def resolve(self) -> List[str]:
        # The dependency names in order
        deps: List[str] = []

        stack = [self.root]

        while len(stack) > 0:
            node = stack.pop()
            has_child_to_resolve = False
            for child in node.children:
                if child.name not in deps:
                    has_child_to_resolve = True
                    stack.append(node)
                    stack.append(child)
                    break

            if not has_child_to_resolve:
                if node.name:
                    deps.append(node.name)

        return deps



import os, os.path
import shutil
from typing import Callable
from enum import Enum

class BuildType(Enum):
    Debug = 0
    Release = 1
BUILD_TYPE_NAMES = {BuildType.Debug: 'debug', BuildType.Release: 'release'}

class StorageController(object):

    def __init__(self, path):
        self.path = path

    def _makedirs(self, p):
        os.makedirs(p, exist_ok=True)
        if os.name == 'nt':
            import ctypes
            def _ensurehidden(p):
                if os.path.basename(p).startswith('.'):
                    FILE_ATTRIBUTE_HIDDEN = 0x02
                    ctypes.windll.kernel32.SetFileAttributesW(p, FILE_ATTRIBUTE_HIDDEN)
            _ensurehidden(p)
            s = os.path.split(p)
            while s[1]:
                _ensurehidden(s[0])
                s = os.path.split(s[0])

    def ensure_directories_exist(self):
        self.walk_directories(lambda p: self._makedirs(p))

    def clean(self):
        self.walk_directories(lambda p: shutil.rmtree(p))

    def walk_directories(self, f: Callable[[str], None]):
        f(self.get_storage_dir())
        f(self.get_sources_dir())
        f(self.get_downloads_dir())
        for t in BuildType:
            f(self.get_builds_dir(t))
            f(self.get_install_dir(t))

    def get_storage_dir(self) -> str:
        return self.join_storage_dir_path()

    def get_sources_dir(self) -> str:
        return self.join_storage_dir_path('sources')

    def get_builds_dir(self, build_type: BuildType) -> str:
        return self.join_storage_dir_path(f'builds-{BUILD_TYPE_NAMES[build_type]}')

    def get_build_dir(self, label: str, build_type: BuildType) -> str:
        return os.path.join(self.get_builds_dir(build_type), label)

    def get_downloads_dir(self) -> str:
        return self.join_storage_dir_path('downloads')

    def get_install_dir(self, build_type: BuildType) -> str:
        return self.join_storage_dir_path(f'prefix-{BUILD_TYPE_NAMES[build_type]}')

    def get_output_prefix_dir(self, build_type: BuildType) -> str:
        return self.join_storage_dir_path(f'output-prefixes-{BUILD_TYPE_NAMES[build_type]}')

    def join_storage_dir_path(self, *args) -> str:
        return os.path.join(self.path, *args)



import copy
import os
import shutil
from typing import Set, Iterable, Dict, List, Optional, Tuple

from dew.projectproperties import ProjectProperties
from dew.dependencygraph import DependencyGraph
from dew.dependencyprocessor import DependencyProcessor
from dew.depstate import DependencyStateController
from dew.dewfile import DewFile, Dependency, ProjectFilesParser
from dew.exceptions import BuildError, DewfileError
from dew.storage import StorageController, BuildType, BUILD_TYPE_NAMES
from dew.view import View


class ProjectProcessor(object):

    def __init__(self, storage: StorageController, properties: ProjectProperties, view: View,
                 depstates: DependencyStateController):
        self.storage = storage
        self.root_dewfile: Optional[DewFile] = None
        self.properties = properties
        self.view = view
        self.depstates = depstates

    def set_data(self, dewfile: DewFile):
        self.root_dewfile = copy.deepcopy(dewfile)

    def process(self):
        dewfile_stack: List[Tuple[DewFile, Optional[str]]] = [(self.root_dewfile, None)]

        # Dependency processors by label
        dependency_processors: Dict[str, DependencyProcessor] = {}

        graph = DependencyGraph()
        deps_needing_build: Set[str] = set()

        while len(dewfile_stack) > 0:
            dewfile, parent_name = dewfile_stack.pop()

            for subdir in dewfile.subdirectories:
                dewfile_path = os.path.join(os.path.dirname(dewfile.path), subdir, 'dewfile.json')
                if os.path.isfile(dewfile_path):
                    parser = ProjectFilesParser(dewfile_path)
                    child_dewfile = parser.parse()
                    dewfile_stack.append((child_dewfile, None))

            for dep in dewfile.dependencies:
                dep_processor = self.make_processor(dep, dewfile)
                label = dep_processor.get_label()
                dependency_processors[label] = dep_processor
                graph.add_dependency(label, parent_name)

                # Connect manual dewfile dependencies
                for dfdep in dep.dependson:
                    graph.add_dependency(dfdep.get_label(), label)

                # if the dependency is up to date, don't bother pulling, as it's already been pulled.
                if not self.depstates.get_state(label):
                    self.view.info('Pulling dependency {0}...'.format(label))
                    dep_processor.pull()
                else:
                    self.view.info(f'Dependency {label} already pulled.')

                if dep_processor.has_dewfile():
                    try:
                        child_dewfile = dep_processor.get_dewfile()
                    except DewfileError as e:
                        self.view.dewfile_error(e)
                        raise

                    dewfile_stack.append((child_dewfile, label))

        for name, processor in dependency_processors.items():
            label = processor.get_label()
            if not self.depstates.get_state(label):
                deps_needing_build.add(label)

        labels_in_order = graph.resolve()

        for label in labels_in_order:
            dep_processor = dependency_processors[label]

            self.view.info('Building dependency {0}...'.format(label))

            node = graph.nodes[label]
            parent_node = node.parent

            while parent_node and parent_node.name:
                self.view.info(f'* which is needed by {parent_node.name}')
                parent_node = parent_node.parent

            if label not in deps_needing_build:
                self.view.info(f'Dependency {label} already built.')
                continue

            # Prepare output prefixes
            child_labels = [n.name for n in node.children]

            for build_type in self.properties.active_build_types():
                self.view.info(f'Building {BUILD_TYPE_NAMES[build_type]}...')
                output_prefix = self.get_isolated_prefix(label, build_type)
                shutil.rmtree(output_prefix)

                input_prefixes = [self.get_isolated_prefix(l, build_type) for l in child_labels]

                # Build and install
                dep_processor.build(output_prefix, input_prefixes, build_type)

            self.depstates.add(label)

        if len(deps_needing_build) > 0:
            for build_type in self.properties.active_build_types():
                self.view.info(f'Updating final {BUILD_TYPE_NAMES[build_type]} prefix')
                self.update_final_prefix(labels_in_order, build_type)

    def make_processor(self, dep: Dependency, dewfile: DewFile) -> DependencyProcessor:
        local_override = dewfile.local_overrides.get(dep.name)
        if local_override:
            dep = copy.deepcopy(dep)
            dep.type = 'local'
            dep.url = local_override
            dep_processor = DependencyProcessor(self.storage, self.view, dep, dewfile, self.properties)
            dep.ref = dep_processor.get_remote().get_latest_ref()

        dep_processor = DependencyProcessor(self.storage, self.view, dep, dewfile, self.properties)
        return dep_processor

    def get_isolated_prefix(self, label: str, type: BuildType) -> str:
        path = os.path.join(self.storage.get_output_prefix_dir(type), label)
        os.makedirs(path, exist_ok=True)
        return path

    def update_final_prefix(self, labels: Iterable[str], build_type: BuildType):
        success = True
        dst_prefix = self.storage.get_install_dir(build_type)
        installed_files: Dict[str, str] = {}

        for label in labels:
            src_prefix = self.get_isolated_prefix(label, build_type)

            for dirpath, dirnames, filenames in os.walk(src_prefix):
                relpath = os.path.relpath(dirpath, src_prefix)
                dst_dir = os.path.join(dst_prefix, relpath)

                for dirname in dirnames:
                    os.makedirs(os.path.join(dst_dir, dirname), exist_ok=True)

                for filename in filenames:
                    src_path = os.path.join(dirpath, filename)
                    dst_path = os.path.join(dst_dir, filename)

                    prefix_neutral_path = os.path.join(relpath, filename)

                    shutil.copy2(src_path, dst_path)

                    if prefix_neutral_path in installed_files:
                        self.view.error(f'Conflicting prefix files found! file: {prefix_neutral_path}, first occurance: {installed_files[prefix_neutral_path]}, current occurance: {label}')
                        success = False
                    else:
                        # Only copy files if we are currently succeding.
                        if success:
                            shutil.copy2(src_path, dst_path)
                        installed_files[prefix_neutral_path] = label

        if not success:
            raise BuildError(f'Failed while installing files')

    def update_refs(self) -> Tuple[DewFile, bool]:
        have_refs_changed = False
        for dep in self.root_dewfile.dependencies:
            if not dep.ref:
                self.view.info(f'Dependency {dep.name} does not have an assigned ref, fetching one now.')
                processor = DependencyProcessor(self.storage, self.view, dep, self.root_dewfile, self.properties)
                dep.ref = processor.get_remote().get_latest_ref()
                have_refs_changed = True
        return self.root_dewfile, have_refs_changed



import os

from typing import Set

from dew.storage import StorageController


class DependencyStateController(object):
    def __init__(self, storage: StorageController) -> None:
        self.storage = storage
        self.states: Set[str] = set()

    def get_state(self, label: str) -> bool:
        return label in self.states

    def load(self) -> None:
        if not os.path.isfile(self.get_state_file_path()):
            return

        with open(self.get_state_file_path()) as f:
            contents = f.read()

        lines = contents.split('\n')

        self.states.clear()

        for line in lines:
            if line:
                self.states.add(line)

    def save(self) -> None:
        contents = '\n'.join(self.states)

        with open(self.get_state_file_path(), 'w') as f:
            f.write(contents)

    def get_state_file_path(self) -> str:
        return self.storage.join_storage_dir_path('depstates')

    def clear(self) -> None:
        self.states.clear()

    def add(self, label: str):
        self.states.add(label)



import os

from dew.builder import Builder
from dew.projectproperties import ProjectProperties
from dew.dewfile import Dependency
from dew.exceptions import BuildError
from dew.subprocesscaller import SubprocessCaller
from dew.view import View
from dew.storage import BuildType

XCODE_BUILD_TYPES = {BuildType.Debug:'Debug', BuildType.Release:'Release'}

class XcodeBuilder(Builder):
    def __init__(self, buildfile_dir: str, build_dir: str, install_dir: str, build_type: BuildType,
                 dependency: Dependency, options: ProjectProperties, caller: SubprocessCaller, view: View) -> None:
        self.buildfile_dir = buildfile_dir
        self.build_dir = build_dir
        self.install_dir = install_dir
        self.build_type = build_type
        self.dependency = dependency
        self.options = options
        self.caller = caller
        self.view = view

    def build(self) -> None:
        build_dir = self.build_dir
        os.makedirs(build_dir, exist_ok=True)

        # Figure out which xcodeproj path we are using
        xcodeproj_path = ''
        for path in os.listdir(os.path.join(self.buildfile_dir)):
            if path.endswith('.xcodeproj'):
                xcodeproj_path = os.path.join(self.buildfile_dir, path)
                break

        if not xcodeproj_path:
            self.view.error('Cannot determine the xcodeproj to build with.')
            raise BuildError()

        self.caller.call(
            [
                'xcodebuild', '-project', xcodeproj_path,
                '-configuration', XCODE_BUILD_TYPES[self.build_type],
                'OBJROOT=' + os.path.join(build_dir, 'Intermediates'),
                'BUILD_DIR=' + os.path.join(build_dir, 'Products'),
                'SYMROOT=' + os.path.join(build_dir, 'Products'),
                'build'
            ],
            cwd=self.build_dir
        )



import argparse
import enum
from typing import Optional


class CommandType(enum.Enum):
    UPDATE = 'update'
    BOOTSTRAP = 'bootstrap'
    CLEAN = 'clean'
    UPGRADE = 'upgrade'
    WORKON = 'workon'
    FINISH = 'finish'
    BUILD = 'build'


class ArgumentData(object):
    def __init__(self) -> None:
        self.help = False
        self.command: Optional[CommandType] = None
        self.verbose = False
        self.version = False
        self.dewfile = 'dewfile.json'
        self.output_path = ''


def make_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--help', '-h', help='Show help information', action='store_true')
    parser.add_argument('command', type=CommandType,
                        help='Command to perform: ' + ', '.join(x.value for x in CommandType))
    parser.add_argument('--verbose', help='Be verbose', dest='verbose', action='store_true')
    parser.add_argument('--version', '-v', action='store_true', help='Output dew version')
    parser.add_argument('--output-path', help='Override the path used for storage of settings and prefixes.')
    parser.add_argument('--dewfile', metavar='DEWFILE_PATH', help='Override the path to the dewfile to use')
    return parser



import sys
import os
import json
from typing import List, Any, Dict

from dew.exceptions import DewfileError, DewError

class Dependency(object):
    def __init__(self) -> None:
        self.name: str = ''
        self.url: str = ''
        self.type: str = ''
        self.head: str = ''
        self.ref: str = ''
        self.buildfile_dir: str = ''
        self.dependson: List[Dependency] = []
        self.cmake_defines: Dict[str, str] = {}

    def get_version(self) -> str:
        return f'{self.type}_{self.ref}'

    def get_label(self) -> str:
        name = self.name
        version = self.get_version()
        return f'{name}_{version}'

class DewFile(object):
    def __init__(self, path: str) -> None:
        self.path = path
        self.subdirectories: List[str] = []
        self.dependencies: List[Dependency] = []
        self.local_overrides: Dict[str, str] = {}

    def find_dependency(self, name: str) -> Dependency:
        try:
            return next(d for d in self.dependencies if d.name == name)
        except Exception as e:
            raise DewfileError(self.path, sys.exc_info()[2], f'"{name}" not found') from e

def _parse_dewfile(data: Dict[str, Any], path: str) -> DewFile:
    subdirectories = []
    if 'subdirectories' in data:
        subdirectories = [str(subdir) for subdir in data['subdirectories']]

    # Ensure subdirectories are valid before proceeding
    for subdir in subdirectories:
        dewfile_path = os.path.join(os.path.dirname(path), subdir, 'dewfile.json')
        if not os.path.isfile(dewfile_path):
            raise DewfileError(path, sys.exc_info()[2], f'subdirectory file "{dewfile_path}" does not exist')

    dependencies = []
    dependencies_obj = []
    if 'dependencies' in data:
        dependencies_obj = data['dependencies']

    # 1st pass: parse dependencies
    for dep in dependencies_obj:
        dependencies.append(parse_dependency(dep))

    dewfile = DewFile(path)
    dewfile.subdirectories = subdirectories
    dewfile.dependencies = dependencies

    # 2nd pass: resolve manual dependencies
    for dep in dewfile.dependencies:
        resolved_deps = []
        for dfdep in dep.dependson:
            resolved_deps.append(dewfile.find_dependency(dfdep))
        dep.dependson = resolved_deps

    return dewfile


def serialize_dewfile(dewfile: DewFile) -> Dict[str, Any]:
    data = {}
    dependencies = []
    for dep in dewfile.dependencies:
        dependencies.append(serialize_dependency(dep))
    if len(dewfile.subdirectories):
        data['subdirectories'] = dewfile.subdirectories
    if len(dependencies):
        data['dependencies'] = dependencies
    return data


def parse_dependency(obj: Dict[str, Any]) -> Dependency:
    dep = Dependency()
    dep.name = obj['name']
    dep.url = obj['url']
    dep.type = obj['type']
    dep.head = obj['head']
    dep.ref = obj['ref']
    dep.buildfile_dir = obj.get('buildfile_dir', '')
    dependson = obj.get('dependson', [])
    if dependson: # Resolved into List[Dependency] in _parse_dewfile
        dep.dependson = [str(d) for d in dependson]
    cmake_defines = obj.get('cmake_defines', {})
    if cmake_defines:
        dep.cmake_defines = {str(k):str(v) for k, v in cmake_defines.items()}
    return dep


def serialize_dependency(dep: Dependency) -> Dict[str, Any]:
    data = {
        'name': dep.name,
        'url': dep.url,
        'type': dep.type,
        'head': dep.head
    }

    if dep.ref:
        data['ref'] = dep.ref
    if dep.buildfile_dir:
        data['buildfile_dir'] = dep.buildfile_dir
    if dep.dependson:
        data['dependson'] = [d.name for d in dep.dependson]
    if dep.cmake_defines:
        data['cmake_defines'] = dep.cmake_defines

    return data


def parse_dewfile(path: str) -> DewFile:
    try:
        dewfile_data = load_json(path)
        dewfile = _parse_dewfile(dewfile_data, path)
    except DewfileError as e:
        raise DewfileError(path, sys.exc_info()[2], e.reason) from e
    except Exception as e:
        raise DewfileError(path, sys.exc_info()[2]) from e
    return dewfile


def parse_local_work_file(path: str) -> Dict[str, str]:
    if not os.path.isfile(path):
        return {}

    with open(path) as f:
        try:
            data = json.load(f)
        except Exception as e:
            raise DewfileError(path, sys.exc_info()[2]) from e

    if not isinstance(data, dict):
        raise DewError('fContents of {path} is misformatted.')

    return {name: str(path) for name, path in data.items()}


def save_dewfile(dewfile: DewFile, path: str) -> None:
    with open(path, 'w') as f:
        json.dump(serialize_dewfile(dewfile), f, indent=4)


def save_local_work(dewfile: DewFile, path: str) -> None:
    with open(path, 'w') as f:
        json.dump(dewfile.local_overrides, f, indent=4)


def load_json(path: str) -> Dict[str, Any]:
    with open(path) as file:
        return json.load(file)


class ProjectFilesParser(object):
    def __init__(self, dewfile_path: str):
        self.dewfile_path = dewfile_path

        path_without_ext, ext = os.path.splitext(self.dewfile_path)
        self.local_work_path = path_without_ext + '.local' + ext

    def parse(self) -> DewFile:
        dewfile = parse_dewfile(self.dewfile_path)
        dewfile.local_overrides = parse_local_work_file(self.local_work_path)
        return dewfile

    def save(self, dewfile: DewFile) -> None:
        save_dewfile(dewfile, self.dewfile_path)

    def save_local_work(self, dewfile: DewFile) -> None:
        save_local_work(dewfile, self.local_work_path)





class Remote(object):
    def pull(self) -> None:
        pass

    def get_latest_ref(self) -> str:
        pass

    def get_current_ref(self) -> str:
        pass

    def get_current_head(self) -> str:
        pass

    def has_pending_changes(self) -> bool:
        pass

    def get_source_dir(self) -> str:
        pass




import os.path
from enum import Enum
from typing import Optional, Iterable

import dew
from dew.builder import Builder
from dew.builder.cmake import CMakeBuilder
from dew.builder.makefile import MakefileBuilder
from dew.builder.xcode import XcodeBuilder
from dew.projectproperties import ProjectProperties
from dew.dewfile import Dependency, DewFile, ProjectFilesParser
from dew.exceptions import BuildError, PullError
from dew.remote import Remote
from dew.remote.git import GitRemote
from dew.remote.local import LocalRemote
from dew.storage import StorageController, BuildType
from dew.subprocesscaller import SubprocessCaller
from dew.view import View


class BuildSystem(Enum):
    UNKNOWN = 0
    CMAKE = 1
    MAKEFILE = 2
    XCODE = 3


class DependencyProcessor(object):
    def __init__(self, storage: StorageController, view: View, dependency: Dependency, dewfile: DewFile,
                 options: ProjectProperties, source_dir: Optional[str] = None):
        self.storage = storage
        self.dependency = dependency
        self.dewfile = dewfile
        self.properties = options
        self.view = view
        self._remote: Optional[Remote] = None
        self.source_dir = source_dir

    def pull(self):
        self.get_remote().pull()

    def build(self, install_dir: str, input_prefixes: Iterable[str], build_type: BuildType):
        self.get_builder(install_dir, input_prefixes, build_type).build()

    def get_remote(self) -> Optional[Remote]:
        if self._remote:
            return self._remote

        type = self.dependency.type
        factory = None

        if type == 'git':
            factory = GitRemote
        elif type == 'local':
            factory = LocalRemote

        if not factory:
            self.view.error('Unknown dependency type {0}'.format(type))
            raise PullError()

        self._remote = factory(
            dependency=self.dependency,
            dest_dir=self.get_default_source_dir()
        )
        return self._remote

    def get_builder(self, install_dir: str, input_prefixes: Iterable[str],
                    build_type: BuildType) -> Optional[Builder]:
        remote = self.get_remote()
        source_dir = remote.get_source_dir()
        buildsystem = self.get_buildsystem(source_dir)
        factory = None

        if buildsystem is BuildSystem.MAKEFILE:
            factory = MakefileBuilder
        elif buildsystem is BuildSystem.CMAKE:
            factory = CMakeBuilder
        elif buildsystem is BuildSystem.XCODE:
            factory = XcodeBuilder

        if not factory:
            self.view.error('Unknown build system')
            raise BuildError()

        caller = SubprocessCaller(self.view)

        return factory(
            buildfile_dir=source_dir,
            build_dir=self.storage.get_build_dir(self.get_label(), build_type),
            install_dir=install_dir,
            additional_prefix_paths=input_prefixes,
            build_type=build_type,
            properties=self.properties,
            caller=caller,
            view=self.view,
            additional_cmake_defines=self.dependency.cmake_defines
        )

    def has_dewfile(self) -> bool:
        dewfile_path = os.path.join(self.get_remote().get_source_dir(), 'dewfile.json')
        return os.path.isfile(dewfile_path)

    def get_dewfile(self) -> Optional[DewFile]:
        dewfile_path = os.path.join(self.get_remote().get_source_dir(), 'dewfile.json')
        if os.path.isfile(dewfile_path):
            parser = ProjectFilesParser(dewfile_path)
            return parser.parse()
        else:
            return None

    def get_default_source_dir(self):
        if self.source_dir:
            return self.source_dir
        return os.path.join(self.storage.get_sources_dir(), self.get_label())

    def get_buildfile_dir(self):
        buildfile_dir = self.dependency.buildfile_dir
        src_dir = self.get_default_source_dir()
        if buildfile_dir:
            return os.path.join(src_dir, buildfile_dir)
        return self.get_default_source_dir()

    def get_buildsystem(self, source_dir: str):
        """ Guesses the build system """
        if os.path.isfile(os.path.join(source_dir, 'CMakeLists.txt')):
            return BuildSystem.CMAKE
        if os.path.isfile(os.path.join(source_dir, 'Makefile')):
            return BuildSystem.MAKEFILE

        for path in os.listdir(source_dir):
            if path.endswith('.xcodeproj'):
                return BuildSystem.XCODE

        return BuildSystem.UNKNOWN

    def install_fake_cmake_config(self):
        with open('', 'w') as f:
            f.write(
                'set(PACKAGE_FIND_NAME "{0}")'
                'set(PACKAGE_FIND_VERSION)'
            )

    def get_version(self) -> str:
        return self.dependency.get_version()

    def get_label(self) -> str:
        return self.dependency.get_label()



import multiprocessing
import os, pathlib, posixpath
from typing import Iterable, ItemsView, Optional

from dew.exceptions import BuildError
from dew.subprocesscaller import SubprocessCaller
from dew.builder import Builder
from dew.projectproperties import ProjectProperties
from dew.view import View
from dew.storage import BuildType

CMAKE_BUILD_TYPES = {BuildType.Debug:'Debug', BuildType.Release:'RelWithDebInfo'}

class CMakeBuilder(Builder):
    def __init__(self,
                 buildfile_dir: str,
                 build_dir: str,
                 install_dir: Optional[str],
                 build_type: BuildType,
                 properties: ProjectProperties,
                 caller: SubprocessCaller,
                 view: View,
                 additional_prefix_paths: Iterable[str] = (),
                 additional_cmake_defines: Optional[ItemsView[str, str]] = None
                 ) -> None:
        self.buildfile_dir = pathlib.PurePath(os.path.abspath(buildfile_dir)).as_posix()
        self.build_dir = pathlib.PurePath(os.path.abspath(build_dir)).as_posix()
        if install_dir:
            self.install_dir = pathlib.PurePath(os.path.abspath(install_dir)).as_posix()
        else:
            self.install_dir = None
        self.build_type = build_type
        self.properties = properties
        self.caller = caller
        self.view = view
        self.additional_prefix_paths = list(additional_prefix_paths)
        self.additional_cmake_defines = additional_cmake_defines

    def get_cmake_executable(self) -> str:
        cmake_executable = self.properties.cmake_executable
        if not cmake_executable:
            cmake_executable = 'cmake'
        return cmake_executable

    def build(self) -> None:
        install_dir = self.install_dir
        os.makedirs(self.build_dir, exist_ok=True)

        cmake_executable = self.get_cmake_executable()

        prefix_paths = self.properties.prefixes.copy()
        prefix_paths.extend(self.additional_prefix_paths)
        prefix_paths = [pathlib.PurePath(prefix).as_posix() for prefix in prefix_paths]

        module_paths = [posixpath.join(prefix, 'share', 'cmake', 'Modules') for prefix in prefix_paths]

        generator = self.properties.cmake_generator
        if not generator:
            generator = guess_generator()
            if not generator:
                raise BuildError('Cannot guess CMake generator to use. ' 
                                 'Please specify a generator using the --cmake-generator flag.')

        args = [
            cmake_executable,
            '-G', generator,
            self.buildfile_dir,
            '-DCMAKE_PREFIX_PATH={0}'.format(';'.join(prefix_paths)),
            '-DCMAKE_MODULE_PATH={0}'.format(';'.join(module_paths)),
            '-DCMAKE_BUILD_TYPE={0}'.format(CMAKE_BUILD_TYPES[self.build_type]),

            # Lots of projects default to build shared libs instead of static. Let's add some consistency.
            # TODO: Need a system to choose the types of libraries desired.
            '-DBUILD_SHARED_LIBS=OFF'
        ]

        if install_dir:
            args.append('-DCMAKE_INSTALL_PREFIX={0}'.format(install_dir),)

        if self.additional_cmake_defines:
            args.extend([f'-D{k}={v}' for k, v in self.additional_cmake_defines.items()])

        # Setup environment
        env = {
            'CC': self.properties.c_compiler_path,
            'CXX': self.properties.cxx_compiler_path,
            'INVOKED_BY_DEW': 'true'  # Set this environment variable to alert the dew CMake modules to no-op.
        }

        # Configure
        self.caller.call(args, cwd=self.build_dir, error_exception=BuildError, env=env)

        build_args = [cmake_executable, '--build', '.']
        if generator in ('Unix Makefiles', 'MinGW Makefiles'):
            build_args.extend(('--', '-j', str(multiprocessing.cpu_count())))

        # Build
        self.caller.call(
            build_args,
            cwd=self.build_dir,
            error_exception=BuildError
        )

        # Install
        if install_dir:
            self.caller.call(
                [cmake_executable, '--build', '.', '--target', 'install'],
                cwd=self.build_dir,
                error_exception=BuildError
            )


GUESSED_GENERATOR: Optional[str] = None


def guess_generator() -> Optional[str]:
    global GUESSED_GENERATOR
    if GUESSED_GENERATOR:
        return GUESSED_GENERATOR

    guess = None

    paths = os.environ.get('PATH', '').split(os.pathsep)
    if os.name == 'nt':
        extensions = [ext.lower() for ext in os.environ.get('PATHEXT').split(';')]
    else:
        extensions = ['']

    def locate(name: str) -> bool:
        for path in paths:
            if not os.path.isdir(path):
                continue
            for entry in os.listdir(path):
                filename, ext = os.path.splitext(entry)
                if filename != name:
                    continue
                if ext.lower() not in extensions:
                    continue
                return True
        return False

    if locate('ninja'):
        guess = 'Ninja'
    elif locate('make'):
        guess = 'Unix Makefiles'
    elif locate('mingw32-make'):
        guess = 'MinGW Makefiles'

    GUESSED_GENERATOR = guess
    return guess



import os
import argparse
from typing import List, Optional

import dew.command
from dew.depstate import DependencyStateController
from dew.impl import CommandData
from dew.projectprocessor import ProjectProcessor
from dew.projectproperties import ProjectProperties
from dew.lockfile import LockFile


class ArgumentData(object):
    def __init__(self) -> None:
        self.cmake_generator = ''
        self.cmake_executable = ''
        self.c_compiler_path = ''
        self.cxx_compiler_path = ''
        self.additional_prefix_paths: List[str] = []
        self.build_type = ''


class Command(dew.command.Command):
    def __init__(self) -> None:
        self.depstates: Optional[DependencyStateController] = None

    def setup_argparser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--CC', dest='c_compiler_path', help='Path to the C compiler')
        parser.add_argument('--CXX', dest='cxx_compiler_path', help='Path to the CXX compiler')
        parser.add_argument('--prefix', dest='additional_prefix_paths', action='append', metavar='PREFIX_PATH')
        parser.add_argument('--cmake-generator', help='The CMake generator to use for dependency projects')
        parser.add_argument('--cmake-executable', help='Path to the CMake executable')
        parser.add_argument('--build-type', help='"debug", "release", or "both"')

    def set_properties_from_args(self, args: ArgumentData, properties: ProjectProperties) -> None:
        if args.cmake_generator:
            properties.cmake_generator = args.cmake_generator
        if args.cmake_executable:
            properties.cmake_executable = args.cmake_executable
        if args.c_compiler_path:
            properties.c_compiler_path = args.c_compiler_path
        if args.cxx_compiler_path:
            properties.cxx_compiler_path = args.cxx_compiler_path
        if args.additional_prefix_paths:
            properties.prefixes = args.additional_prefix_paths
        if args.build_type:
            properties.build_type = args.build_type

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        with LockFile(data.storage.join_storage_dir_path('lock'), data):
            dewfile = data.project_parser.parse()
            depstates = DependencyStateController(data.storage)
            self.depstates = depstates
            depstates.load()

            is_dirty = data.properties.dirty
            if is_dirty:
                data.view.info('Dew project properties have changed. Clearing dependency state information.')
                depstates.clear()
                depstates.save()

            data.properties.save()
            properties = data.properties.get()

            project_processor = ProjectProcessor(data.storage, properties, data.view, depstates)
            project_processor.set_data(dewfile)

            dewfile, refs_have_changed = project_processor.update_refs()
            if refs_have_changed:
                data.project_parser.save_refs(dewfile)

            project_processor.process()

        return 0

    def cleanup(self, args: ArgumentData, data: CommandData) -> None:
        if self.depstates:
            self.depstates.save()



import fasteners, threading
from dew.impl import CommandData

class LockFile(fasteners.InterProcessLock):
    def __init__(self, path: str, data: CommandData):
        self.data = data
        super().__init__(path)

    def __enter__(self):
        gotten = super().acquire(blocking=False)
        if not gotten:
            self.data.view.info("Waiting for another dew to complete."
                                " Press Ctrl-C to abort.")
            gotten = super().acquire()
            if not gotten:
                # This shouldn't happen, but just incase...
                raise threading.ThreadError("Unable to acquire a file lock"
                                            " on `%s` (when used as a"
                                            " context manager)" % self.path)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().release()



import sys

import argparse
import os
from typing import Optional, List

import dew.args
from dew.args import ArgumentData, CommandType
from dew.command import Command
from dew.dewfile import ProjectFilesParser
from dew.exceptions import DewfileError, DewError
from dew.projectproperties import ProjectPropertiesController
from dew.impl import CommandData
from dew.storage import StorageController
from dew.view import View
import dew.git
import dew.commands.update
import dew.commands.bootstrap
import dew.commands.clean
import dew.commands.upgrade
import dew.commands.workon
import dew.commands.finish
import dew.commands.build


# TODO: Import these dynamically
command_module_map = {
    CommandType.UPDATE: dew.commands.update,
    CommandType.BOOTSTRAP: dew.commands.bootstrap,
    CommandType.CLEAN: dew.commands.clean,
    CommandType.UPGRADE: dew.commands.upgrade,
    CommandType.WORKON: dew.commands.workon,
    CommandType.FINISH: dew.commands.finish,
    CommandType.BUILD: dew.commands.build
}


def main() -> int:
    parser = dew.args.make_argparser()
    args = ArgumentData()

    # If there is a help flag defined, figure out what the positional argument is. Turns out argparse is kinda lame.
    argv: List[str] = sys.argv.copy()
    del argv[0]
    is_help = False
    positional: Optional[str] = None

    if '-h' in argv or '--help' in argv:
        is_help = True
        is_parsing_param = False
        for arg in argv:
            if is_parsing_param:
                continue
            if len(arg) > 0:
                if arg[0] == '-':
                    if len(arg) == 1 or arg[1] == '-':
                        is_parsing_param = True
                    continue

            positional = arg
            break

    if '--version' in argv or '-v' in argv:
        print(f'dew {dew.VERSION_MAJOR}.{dew.VERSION_MINOR}.{dew.VERSION_PATCH}')
        return 0

    view = View()

    if is_help:
        help(view, positional, parser)
        return 1

    # noinspection PyTypeChecker
    _, remaining_args = parser.parse_known_args(namespace=args)

    view.show_verbose = True if args.verbose else False

    # Default command
    if not args.command:
        args.command = CommandType.UPDATE

    command_module = command_module_map.get(args.command)
    command = command_module.Command()

    if command_module is None:
        view.error(f'Don\'t know how to handle command {args.command.value}! This is a bug.')
        return 1

    command_args = command_module.ArgumentData()
    command_argparser = get_command_argparser(args.command.value, command)
    # noinspection PyTypeChecker
    command_argparser.parse_args(remaining_args, namespace=command_args)

    storage = get_storage(args)
    properties = get_properties(storage, command_args, command)
    project_parser = ProjectFilesParser(args.dewfile)

    command_data = CommandData(args, view, storage, properties, project_parser)

    try:
        return command.execute(command_args, command_data)
    except DewfileError as e:
        view.dewfile_error(e)
        return 1
    except DewError as e:
        view.error(f'Dew did not finish successfully :(\n{e}')
        return 1
    finally:
        command.cleanup(command_args, command_data)


def help(view: View, command_name: Optional[str], parser: argparse.ArgumentParser) -> None:
    command_type = None
    if command_name is not None:
        for e in CommandType:
            if e.value == command_name:
                command_type = e
                break

        if command_type is None:
            view.error(f'{command_name} is not a command!')
            return

    if command_type is not None:
        command_module = command_module_map.get(command_type)
        command = command_module.Command()
        command_parser = get_command_argparser(command_name, command)
        module_help(view, command_name, command_parser)
        return

    parser.print_help()
    view.info('\nSpecify a command with the help flag for help on a specific command.')


def get_command_argparser(command_name: str, command: Command) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog=f'dew {command_name}', add_help=False)
    command.setup_argparser(parser)
    return parser


def module_help(view: View, name: str, parser: argparse.ArgumentParser) -> None:
    parser.print_help()


def get_storage(args: ArgumentData) -> StorageController:
    if args.output_path:
        storage_path = args.output_path
    else:
        storage_path = os.path.join(os.getcwd(), '.dew')

    storage = StorageController(storage_path)
    storage.ensure_directories_exist()
    return storage


def get_properties(storage: StorageController, command_args, command: Command) -> ProjectPropertiesController:
    controller = ProjectPropertiesController(storage)
    controller.load()
    properties = controller.get()
    command.set_properties_from_args(command_args, properties)
    controller.set(properties)
    return controller


def main_with_exit() -> None:
    exit(main())



import copy
import json
import os.path
from typing import Dict, Union, Any, List, Tuple

from dew.storage import StorageController, BuildType

BUILD_TYPE_TUPLES = {'debug':(BuildType.Debug,),
                     'release':(BuildType.Release,),
                     'both':(BuildType.Debug, BuildType.Release)}

class ProjectProperties(object):
    def __init__(self):
        self.cmake_generator = ''
        self.cmake_executable = ''
        self.c_compiler_path = ''
        self.cxx_compiler_path = ''
        self.prefixes: List[str] = []
        self.options: Dict[str, Union[str, bool]] = {}
        self.build_type = 'both'

    def active_build_types(self) -> Tuple[BuildType]:
        return BUILD_TYPE_TUPLES[self.build_type]

class ProjectPropertiesController(object):
    def __init__(self, storage: StorageController):
        self.storage = storage
        self.properties_dict: Dict[str, Any] = {}
        self.dirty = False

    def load(self) -> None:
        cache_path = self.get_cache_file_path()
        if os.path.isfile(cache_path):
            with open(cache_path) as f:
                properties = self.from_dict(json.load(f))
        else:
            properties = ProjectProperties()
        self.properties_dict = self.to_dict(properties)
        self.dirty = False

    def save(self) -> None:
        cache_path = self.get_cache_file_path()
        with open(cache_path, 'w') as f:
            json.dump(self.properties_dict, f, indent=4)
        self.dirty = False

    def get(self) -> ProjectProperties:
        return self.from_dict(self.properties_dict)

    def set(self, properties: ProjectProperties) -> None:
        new_dict = self.to_dict(properties)
        if self.are_dicts_different(self.properties_dict, new_dict):
            self.dirty = True
        self.properties_dict = new_dict

    def from_dict(self, data: Dict[str, Any]) -> ProjectProperties:
        properties = ProjectProperties()
        properties.cmake_generator = data.get('cmake_generator', '')
        properties.cmake_executable = data.get('cmake_executable', '')
        properties.c_compiler_path = data.get('c_compiler_path', '')
        properties.cxx_compiler_path = data.get('cxx_compiler_path', '')
        properties.prefixes = list(data.get('prefixes', []))
        properties.options = dict(data.get('options', {}))
        properties.build_type = data.get('build_type', 'both')
        if properties.build_type not in ('debug', 'release', 'both'):
            properties.build_type = 'both'
        return properties

    def to_dict(self, properties: ProjectProperties) -> Dict[str, str]:
        data = dict()
        data['cmake_generator'] = properties.cmake_generator
        data['cmake_executable'] = properties.cmake_executable
        data['c_compiler_path'] = properties.c_compiler_path
        data['cxx_compiler_path'] = properties.cxx_compiler_path
        data['prefixes'] = properties.prefixes.copy()
        data['options'] = copy.deepcopy(properties.options)
        data['build_type'] = properties.build_type
        return data

    def get_cache_file_path(self) -> str:
        return os.path.join(self.storage.get_storage_dir(), 'cache.json')

    def are_objects_different(self, object_a: Any, object_b: Any) -> bool:
        if type(object_a) != type(object_b):
            return True

        if isinstance(object_a, dict):
            return self.are_dicts_different(object_a, object_b)
        if isinstance(object_a, list):
            return self.are_lists_different(object_a, object_b)

        return object_a != object_b

    def are_lists_different(self, list_a: List[Any], list_b: List[Any]) -> bool:
        if len(list_a) != len(list_b):
            return True

        for i in range(len(list_a)):
            value_a = list_a[i]
            value_b = list_b[i]
            if self.are_objects_different(value_a, value_b):
                return True

        return False

    def are_dicts_different(self, dict_a: Dict[str, Any], dict_b: Dict[str, Any]) -> bool:
        for name, value in dict_a.items():
            other_value = dict_b.get(name)

            if self.are_objects_different(value, other_value):
                return True

        return False



import argparse
import os

import dew.command
from dew.builder.cmake import CMakeBuilder
from dew.impl import CommandData
from dew.subprocesscaller import SubprocessCaller


class ArgumentData(object):
    def __init__(self) -> None:
        self.build_directory = ''


class Command(dew.command.Command):
    def setup_argparser(self, parser: argparse.ArgumentParser):
        parser.add_argument('build_directory', metavar='BUILD_DIRECTORY', help='Path to the build directory')

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        caller = SubprocessCaller(data.view, redirect_output=False)

        builder = CMakeBuilder(
            buildfile_dir=os.getcwd(),
            build_dir=args.build_directory,
            install_dir=None,
            properties=data.properties.get(),
            caller=caller,
            view=data.view,
            additional_prefix_paths=(data.storage.get_install_dir(), )
        )

        builder.build()
        return 0



import subprocess, os
from typing import List, Type, Optional, Dict

from dew.view import View


class SubprocessCallError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class SubprocessCaller(object):
    def __init__(self, view: View, *, redirect_output=True) -> None:
        self.view = view
        self.redirect_output = redirect_output

    def call(self,
             args: List[str],
             cwd: str,
             error_exception: Type[Exception],
             env: Optional[Dict[str, str]] = None
             ) -> None:
        if env is not None:
            merged_env = dict(os.environ)
            merged_env.update(env)
        else:
            merged_env = os.environ
        self.view.verbose('Calling subprocess: "{0}", cwd: {1}'.format(repr(args), repr(cwd)))

        output_file = subprocess.PIPE
        if not self.redirect_output:
            output_file = None

        proc = subprocess.run(
            args, cwd=cwd, stdout=output_file, stderr=output_file, env=merged_env,
            universal_newlines=True
        )

        self.view.verbose(f'Process output:\n{proc.stdout}')
        self.view.verbose(f'Process stderr:\n{proc.stderr}')

        if proc.returncode is not 0:
            if len(proc.stderr) > 0:
                self.view.error(proc.stderr)
            raise error_exception()
